#ifndef __NAVBASE__HH_
#define __NAVBASE__HH_ 

// Navigation Class

#include <sys/time.h>

#include <Performer/pf/pfNode.h>
#include <Performer/pf/pfSCS.h>
#include <Performer/pf/pfDCS.h>

//==========>NavBase<====================================
/*------------------------------------------------
class:   NavBase
purpose: Base Class for Cave Navigation using Performer
         Handles basic rotations and translations in cave space.
	  The most important inherited methods are
	 update();  // do all your movement computations
	            // for the nav_dcs
        updateDCS();  // commit the changes done in the update
	               // to the DCS..  This may later become
		       // embedded into the update() methods
		       // of the inherited classes for convenience.
-------------------------------------------------*/
class NavBase {
protected:
  pfDCS *navDCS;
public:
  // the below transformations are relative to the current transformation
  // state (eg. they are cumulative.  To make them absolute transformations, 
  // simply loadIdentity() before you call them.
  virtual void rotate(float rads,char axis); // rotate around axis 'x','y',or 'z'
  virtual void rotate(float *r);  // r is an array with values of rotation
                // in radians around 'x','y', and 'z' respectively
  virtual void translate(float x,float y,float z);  // translate in x,y & z
  virtual void translate(float *v); // same as above but using an array
  virtual void loadIdentity();      // undo all navigation transformations
  
  NavBase(pfDCS *nav); // constructor, stores navDCS internally
  virtual void updateDCS();    // commit the transformations to the navDCS
  virtual void update();  // compute the transformations (rotation/translation)
  virtual void reset();   // aliased to loadIdentity()... purpose is to
                          // reset the class to the initial conditions 
                          // of when it was constructed (so you don't need to
                          // destroy and re-construct it.
};


//==========>NavModel<====================================
/*------------------------------------------------
class:   NavModel
purpose: Base Class for Model Navigation
         Encapsulates basic rotations and translations 
	 for DCS's in a manner that is interchangable
	 with the NavBase methods.  This allows you to
	 do CaveNav like transformations on individual
	 models in the scene.  So just copy the classes
	 that inherit from NavBase and drop them into
	 NavModel in order use it.

	 The most important inherited methods are
	 update();  // do all your movement computations
	            // for the nav_dcs
         updateDCS();  // commit the changes done in the update
	               // to the DCS..  This may later become
		       // embedded into the update() methods
		       // of the inherited classes for convenience.
-------------------------------------------------*/
class NavModel : public NavBase{
  float trans[3],rot[3];
public:
  // the below transformations are relative to the current transformation
  // state (eg. they are cumulative.  To make them absolute transformations, 
  // simply loadIdentity() before you call them.
  virtual void rotate(float rads,char axis); // rotate around axis 'x','y',or 'z'
  virtual void rotate(float *r);  // r is an array with values of rotation
                // in radians around 'x','y', and 'z' respectively
  virtual void translate(float x,float y,float z);  // translate in x,y & z
  virtual void translate(float *v); // same as above but using an array
  virtual void loadIdentity();      // undo all navigation transformations
  
  NavModel(pfDCS *nav); // constructor, stores navDCS internally
  virtual void updateDCS();    // commit the transformations to the navDCS
  virtual void update();  // compute the transformations (rotation/translation)
  virtual void reset();   // aliased to loadIdentity()... purpose is to
                          // reset the class to the initial conditions 
                          // of when it was constructed (so you don't need to
                          // destroy and re-construct it.
};
//==========>NavRate<====================================
/*------------------------------------------------------
class: NavRate
purpose: Base Class for Navigation forms that define movement
         in terms of velocity instead of absolute coordinates.
	 The most important methods in this class are
	 updateTime(); // recomputes the time change since the
	               // last update so it can determine
		       // velocity.
         rotationVelocity()/translationVelocity()
------------------------------------------------------*/
class NavRate : public NavBase{
protected:
  //long lastTime,thisTime; // in uSec
  double lastTime,thisTime;
  //float dt;
  double dt;
  // rotate at a specified velocity (radians/second)
  void rotationVelocity(float rads_sec,char axis);
  // translate at a specified velocity (coordinates/sec)
  void translationVelocity(float vx,float vy,float vz);
  void updateTime();
public:
  NavRate(pfDCS *dcs);
  virtual void update(); // just calls updateTime().. (just a placeholder)
  virtual void reset(); // resets the timer (not much else)
};

//==========>NavFly<====================================
/*------------------------------------------------------
Class: NavFly
Purpose: Implements the ever-so-common "Cave Fly Mode".
         The only advantage over the traditional method
	 is that all of the guts and relevant data is
	 encapsulated so you need only call "update()"
	 to go flying around your scene.

	 JOYSTICK_X: Controls turning from left to right.
	             (eg rotation around the 'z' axis)
         JOYSTICK_Y: Controls forward velocity.  You will
	             fly in the direction the wand is pointing
Important Methods:
         The key methods in here are setting the
	 Translation Speed and the Rotation speed
	 using (setTranslationSpeed() and setRotationSpeed())
	 The update() method handles the rest by monitoring
	 the joystick values and multiplying by the speed
	 factors.
------------------------------------------------------*/
class NavFly : public NavRate {
protected:
  float rSpeed,tSpeed;  
public:
  NavFly(pfDCS *nav);
  virtual void update();
  void setRotationSpeed(float speed); // where speed is in rad/sec
  void adjustRotationSpeed(float factor); // adjust speed by factor (speed*=factor)
  void setTranslationSpeed(float speed); // where speed is the trans speed in clicks/sec
  void adjustTranslationSpeed(float factor); // adjust speed by factor (speed*=factor)
  void purgeTime(){
    updateTime();
    updateTime();
  }
};

// class NavGrab??  grab() ungrab()
//==========>NavGrabAbsolute<============================
/*
  Description: Grab an object with the wand.
  Its position follows the wand position and its orientation
  follows the wand orientation.  The object (regardless of location)
  will snap to the wands position and orientation when the grab
  starts.
  It grabs on instantiation.
  To ungrab, destroy the NavGrab object.
*/
class NavGrabAbsolute : public NavBase {
  void negateVF3(float *v);
public:
  NavGrabAbsolute(pfDCS *dcs);
  virtual void update();
};

//==========>NavGrabRelative<============================
/* 
   Description: Grab an object with the wand.
   Unlike the NavGrabAbsolute, this version grabs
   relative to the wand's initial orientation and
   position and only tracks changes in that position.
*/
class NavGrabRelative : public NavBase {
  void negateVF3(float *v);
  float initialWandOrientation[3];
  float initialWandPosition[3];
  
  void offsetV3(float v1[3],float v2[3],float v[3]);
public:
  NavGrabRelative(pfDCS *dcs);
  virtual void update();
  virtual void reset();
};

//==========>NavWalk<=================================
/*
  We'll fill in the guts of this one by cut-n-paste from
  the stock-exchange stuff that Ed is working with.
*/
typedef pfNode *pfNodePtr;
class NavWalk : public NavFly{  
  pfNodePtr &model; // The scene graph to test intersection with

  // intersection testing equipment
  pfSegSet segset;
  boolean intersection_caching;
  int samplerate,nsamples;
  float height,maxFallVelocity;
  double velocity,acceleration,lastTime;
  void computeNavHeight();
public:
  NavWalk(pfDCS *navdcs,pfNodePtr &mdl,int icache=1);
  virtual void update();
  inline void setHeight(float h){
    height=h;
  }
  inline void setAcceleration(float a){
    acceleration=a;
  }
  inline void setMaxFallVelocity(float v){
    maxFallVelocity=v;
  }
  inline void setSampleRate(int rate){
    samplerate=rate;
  }
  inline int getSampleRate(){
    return samplerate;
  }
  void intersectionCaching(boolean icache=1);
};

#endif
