#ifndef __READATOMFILES_HH_
#define __READATOMFILES_HH_

// Convenience structures for locating a point in 3D space
struct sPoint { float x,y,z; };
union Point {
  float array[3];
  sPoint cart;
};

// Definition of an individual atom.
struct AtomDef {
  int type; // Integer type for the atom (generated from name by AtomType class)
  Point p;  // Location of the center of the atom.
};

// Brute-force name manager for atom namestrings
// It converts the name strings into index numbers
// and vice-verse.
class AtomType {
  // typedef char AtomTypeDefn[32];
  struct AtomTypeDefn {
    char name[32];
    // Default (empty) constructor.
    AtomTypeDefn(){}
    // Copy Constructor
    AtomTypeDefn(AtomTypeDefn &a){
      strcpy(name,a.name);
    }
    // Defines behavior of the '=' operation for assignment/
    AtomTypeDefn &operator=(AtomTypeDefn &a){
      if(&a==this) return *this;
      strcpy(name,a.name);
      return *this;
    }
  };
  FlexArray<AtomTypeDefn> atoms;
public:
  AtomType(){}
  AtomType(AtomType &at);
  // Converts a type name string into a unique integer number.
  // On first encounter with the string, it will create a new number for
  // it.  On subsequent encounters with the same string, it will 
  // return the same number (uniquely identifying that string).
  int getTypeNumber(char *name);
  // retrieves the typename given the type number
  inline char *getTypeName(int typenumber){
    return (atoms[typenumber]).name;
  }
  // Gets the number of types currently identified (number of 
  // uniquely identifiable strings).
  inline int nTypes(){ return atoms.getSize(); }
  // copy operator '='.
  AtomType &operator=(AtomType &at);
};

// A list of atoms (basically all of the atoms in the lattice at a particular
// timestep). 
class AtomList {
public:
  AtomType atom_types;
  FlexArray<AtomDef> atoms;
  AtomList(){}
  // copy constructor (for copying AtomDef Lists)
  AtomList(AtomList &a);
  // copy operator (for copying AtomDef Lists using "=")
  AtomList &operator=(AtomList &a);
  AtomList(char *filename);
  // Read the atom locations from an x,y,z file named <filename>
  void read(char *filename);
  void print(); // debug print.
};

// Basically a time-sequence of AtomLists.  This linearly interpolates
// between the AtomLists to create smooth animations of the time-evolution
// of atom positions for the Cave application.
class TimeAtomList {
  FlexArray<AtomList> timesteps;
  static char *generateName(char *buffer,char *basename,int index);
public:
  AtomType atom_types; // copied from 0th timestep (for convenience)
  TimeAtomList(char *namebase,int minindex,int maxindex);
  // destructor (called when you "delete" this object)
  ~TimeAtomList();
  // Gets the minimum and maximum indices for timesteps
  // The timestep will generally range from 0 to the number of timesteps-1.
  void getRange(float &min,float &max);
  // linearly interpolate between timesteps
  // based on a floating-point "time" value
  // put the results in the array "atoms"
  void getAtoms(float time,FlexArray<AtomDef> &atoms);
};

void print(FlexArray<AtomDef> &atoms); // Debug print statement

#endif
